from unittest import TestCase

import inject
from inject import autoparams, ConstructorTypeError


class TestFunctional(TestCase):
    def tearDown(self):
        inject.clear()

    def test(self):
        class Config(object):
            def __init__(self, greeting):
                self.greeting = greeting

        class Cache(object):
            config = inject.attr(Config)

            def load_greeting(self):
                return self.config.greeting

        class User(object):
            cache = inject.attr(Cache)

            def __init__(self, name):
                self.name = name

            def greet(self):
                return '%s, %s' % (self.cache.load_greeting(), self.name)

        def config(binder):
            binder.bind(Config, Config('Hello'))

        inject.configure(config)

        user = User('John Doe')
        greeting = user.greet()
        assert greeting == 'Hello, John Doe'

    def test_class_with_restricted_bool_casting(self):
        class DataFrame(object):
            def __nonzero__(self):
                """Python 2"""
                raise NotImplementedError('Casting to boolean is not allowed')

            def __bool__(self):
                """Python 3"""
                raise NotImplementedError('Casting to boolean is not allowed')

        def create_data_frame():
            return DataFrame()

        def config(binder):
            binder.bind_to_constructor(DataFrame, create_data_frame)

        inject.configure(config)

        assert type(inject.instance(DataFrame)) is DataFrame
        # There should not be an error when we fetch created instance
        assert type(inject.instance(DataFrame)) is DataFrame

    def test_class_support_in_autoparams_programmaticaly(self):
        class AnotherClass:
            pass

        class SomeClass:
            def __init__(self, another_object: AnotherClass):
                self.another_object = another_object

        def config(binder):
            binder.bind_to_constructor(SomeClass, autoparams()(SomeClass))
            binder.bind_to_constructor(AnotherClass, autoparams()(AnotherClass))

        inject.configure(config)

        some_object = inject.instance(SomeClass)
        assert type(some_object) is SomeClass
        assert type(some_object.another_object) is AnotherClass

    def test_error_message(self):
        class SomeClass:
            def __init__(self, missing_arg):
                self._missing_arg = missing_arg

        class AnotherClass(SomeClass):
            pass

        def create_some_class(missing_arg):
            return AnotherClass(missing_arg)

        def config(binder):
            binder.bind_to_constructor(SomeClass, autoparams()(SomeClass))
            binder.bind_to_constructor(AnotherClass, create_some_class)

        inject.configure()

        # covers case when no constructor provided
        try:
            inject.instance(SomeClass)
        except ConstructorTypeError as err:
            assert 'SomeClass' in str(err)
            assert 'missing_arg' in str(err)

        inject.clear_and_configure(config)

        # covers case with provided constructor
        try:
            inject.instance(SomeClass)
        except ConstructorTypeError as err:
            assert 'SomeClass' in str(err)
            assert 'missing_arg' in str(err)

        try:
            inject.instance(AnotherClass)
        except TypeError as err:
            assert not isinstance(err, ConstructorTypeError)
            assert 'create_some_class' in str(err)
            assert 'missing_arg' in str(err)


from unittest import TestCase
import asyncio
import inject


class BaseTestInject(TestCase):
    def tearDown(self):
        inject.clear()
    
    def run_async(self, awaitable):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(awaitable)


import inject
from test import BaseTestInject


class TestInjectAttr(BaseTestInject):
    def test_attr(self):
        class MyClass(object):
            field = inject.attr(int)

        inject.configure(lambda binder: binder.bind(int, 123))
        my = MyClass()
        value0 = my.field
        value1 = my.field

        assert value0 == 123
        assert value1 == 123

    def test_class_attr(self):
        class MyClass(object):
            field = inject.attr(int)

        inject.configure(lambda binder: binder.bind(int, 123))
        value0 = MyClass.field
        value1 = MyClass.field

        assert value0 == 123
        assert value1 == 123


class TestInjectAttrDataclass(BaseTestInject):
    def test_class_attr_dc(self):
        class MyClass(object):
            field = inject.attr_dc(int)

        inject.configure(lambda binder: binder.bind(int, 123))
        value0 = MyClass.field
        value1 = MyClass.field

        assert value0 == 123
        assert value1 == 123




import sys
from typing import Optional

from test import BaseTestInject

import inject


class A: pass
class B: pass
class C: pass


class TestInjectEmptyAutoparamsWithBraces(BaseTestInject):
    @staticmethod
    def _get_decorator():
        return inject.autoparams()

    def test_autoparams_by_class(self):
        @self._get_decorator()
        def test_func(val: int = None):
            return val

        inject.configure(lambda binder: binder.bind(int, 123))

        assert test_func() == 123
        assert test_func(val=321) == 321

    def test_autoparams_multi(self):
        @self._get_decorator()
        def test_func(a: A, b: B, *, c: C):
            return a, b, c

        def config(binder):
            binder.bind(A, 1)
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)

        assert test_func() == (1, 2, 3)
        assert test_func(10) == (10, 2, 3)
        assert test_func(10, 20) == (10, 20, 3)
        assert test_func(10, 20, c=30) == (10, 20, 30)
        assert test_func(a='a') == ('a', 2, 3)
        assert test_func(b='b') == (1, 'b', 3)
        assert test_func(c='c') == (1, 2, 'c')
        assert test_func(a=10, c=30) == (10, 2, 30)
        assert test_func(c=30, b=20, a=10) == (10, 20, 30)
        assert test_func(10, b=20) == (10, 20, 3)

    def test_autoparams_strings(self):
        @self._get_decorator()
        def test_func(a: 'A', b: 'B', *, c: 'C'):
            return a, b, c

        def config(binder):
            binder.bind(A, 1)
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)

        assert test_func() == (1, 2, 3)
        assert test_func(10) == (10, 2, 3)
        assert test_func(10, 20) == (10, 20, 3)
        assert test_func(10, 20, c=30) == (10, 20, 30)
        assert test_func(a='a') == ('a', 2, 3)
        assert test_func(b='b') == (1, 'b', 3)
        assert test_func(c='c') == (1, 2, 'c')
        assert test_func(a=10, c=30) == (10, 2, 30)
        assert test_func(c=30, b=20, a=10) == (10, 20, 30)
        assert test_func(10, b=20) == (10, 20, 3)

    def test_autoparams_with_defaults(self):
        @self._get_decorator()
        def test_func(a=1, b: 'B' = None, *, c: 'C' = 300):
            return a, b, c

        def config(binder):
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)

        assert test_func() == (1, 2, 3)
        assert test_func(10) == (10, 2, 3)
        assert test_func(10, 20) == (10, 20, 3)
        assert test_func(10, 20, c=30) == (10, 20, 30)
        assert test_func(a='a') == ('a', 2, 3)
        assert test_func(b='b') == (1, 'b', 3)
        assert test_func(c='c') == (1, 2, 'c')
        assert test_func(a=10, c=30) == (10, 2, 30)
        assert test_func(c=30, b=20, a=10) == (10, 20, 30)
        assert test_func(10, b=20) == (10, 20, 3)

    def test_autoparams_on_method(self):
        class Test:
            @self._get_decorator()
            def func(self, a=1, b: 'B' = None, *, c: 'C' = None):
                return self, a, b, c

        def config(binder):
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)
        test = Test()

        assert test.func() == (test, 1, 2, 3)
        assert test.func(10) == (test, 10, 2, 3)
        assert test.func(10, 20) == (test, 10, 20, 3)
        assert test.func(10, 20, c=30) == (test, 10, 20, 30)
        assert test.func(a='a') == (test, 'a', 2, 3)
        assert test.func(b='b') == (test, 1, 'b', 3)
        assert test.func(c='c') == (test, 1, 2, 'c')
        assert test.func(a=10, c=30) == (test, 10, 2, 30)
        assert test.func(c=30, b=20, a=10) == (test, 10, 20, 30)
        assert test.func(10, b=20) == (test, 10, 20, 3)

    def test_autoparams_on_classmethod(self):
        class Test:
            # note inject must be *before* classmethod!
            @classmethod
            @self._get_decorator()
            def func(cls, a=1, b: 'B' = None, *, c: 'C' = None):
                return cls, a, b, c

        def config(binder):
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)

        assert Test.func() == (Test, 1, 2, 3)
        assert Test.func(10) == (Test, 10, 2, 3)
        assert Test.func(10, 20) == (Test, 10, 20, 3)
        assert Test.func(10, 20, c=30) == (Test, 10, 20, 30)
        assert Test.func(a='a') == (Test, 'a', 2, 3)
        assert Test.func(b='b') == (Test, 1, 'b', 3)
        assert Test.func(c='c') == (Test, 1, 2, 'c')
        assert Test.func(a=10, c=30) == (Test, 10, 2, 30)
        assert Test.func(c=30, b=20, a=10) == (Test, 10, 20, 30)
        assert Test.func(10, b=20) == (Test, 10, 20, 3)

    def test_autoparams_on_classmethod_ob_object(self):
        class Test:
            # note inject must be *before* classmethod!
            @classmethod
            @self._get_decorator()
            def func(cls, a=1, b: 'B' = None, *, c: 'C' = None):
                return cls, a, b, c

        def config(binder):
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)
        test = Test

        assert test.func() == (Test, 1, 2, 3)
        assert test.func(10) == (Test, 10, 2, 3)
        assert test.func(10, 20) == (Test, 10, 20, 3)
        assert test.func(10, 20, c=30) == (Test, 10, 20, 30)
        assert test.func(a='a') == (Test, 'a', 2, 3)
        assert test.func(b='b') == (Test, 1, 'b', 3)
        assert test.func(c='c') == (Test, 1, 2, 'c')
        assert test.func(a=10, c=30) == (Test, 10, 2, 30)
        assert test.func(c=30, b=20, a=10) == (Test, 10, 20, 30)
        assert test.func(10, b=20) == (Test, 10, 20, 3)

    def test_autoparams_omits_return_type(self):
        @self._get_decorator()
        def test_func(a: str) -> int:
            return a

        def config(binder):
            binder.bind(str, 'bazinga')

        inject.configure(config)

        assert test_func() == 'bazinga'


class TestInjectEmptyAutoparamsNoBraces(TestInjectEmptyAutoparamsWithBraces):
    @staticmethod
    def _get_decorator():
        return inject.autoparams


class TestInjectSelectedAutoparams(BaseTestInject):

    def test_autoparams_only_selected(self):
        @inject.autoparams('a', 'c')
        def test_func(a: 'A', b: 'B', *, c: 'C'):
            return a, b, c

        def config(binder):
            binder.bind(A, 1)
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)

        self.assertRaises(TypeError, test_func)
        self.assertRaises(TypeError, test_func, a=1, c=3)

    def test_autoparams_only_selected_with_optional(self):
        @inject.autoparams('a', 'c')
        def test_func(a: 'A', b: 'B', *, c: Optional[C] = None):
            return a, b, c

        def config(binder):
            binder.bind(A, 1)
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)

        self.assertRaises(TypeError, test_func)
        self.assertRaises(TypeError, test_func, a=1, c=3)

    def test_autoparams_only_selected_with_optional_pep604_union(self):
        if not sys.version_info[:3] >= (3, 10, 0):
            return

        @inject.autoparams('a', 'c')
        def test_func(a: 'A', b: 'B', *, c: C | None = None):
            return a, b, c

        def config(binder):
            binder.bind(A, 1)
            binder.bind(B, 2)
            binder.bind(C, 3)

        inject.configure(config)

        self.assertRaises(TypeError, test_func)
        self.assertRaises(TypeError, test_func, a=1, c=3)



from unittest import TestCase

from inject import Binder, InjectorException


class TestBinder(TestCase):
    def test_bind(self):
        binder = Binder()
        binder.bind(int, 123)

        assert int in binder._bindings

    def test_bind__class_required(self):
        binder = Binder()

        self.assertRaisesRegex(InjectorException, 'Binding key cannot be None', binder.bind, None, None)

    def test_bind__duplicate_binding(self):
        binder = Binder()
        binder.bind(int, 123)

        self.assertRaisesRegex(InjectorException, "Duplicate binding", binder.bind, int, 456)

    def test_bind__allow_override(self):
        binder = Binder(allow_override=True)
        binder.bind(int, 123)
        binder.bind(int, 456)
        assert int in binder._bindings

    def test_bind_provider(self):
        provider = lambda: 123
        binder = Binder()
        binder.bind_to_provider(int, provider)

        assert binder._bindings[int] is provider

    def test_bind_provider__provider_required(self):
        binder = Binder()
        self.assertRaisesRegex(InjectorException, "Provider cannot be None", binder.bind_to_provider, int, None)

    def test_bind_constructor(self):
        constructor = lambda: 123
        binder = Binder()
        binder.bind_to_constructor(int, constructor)

        assert binder._bindings[int]._constructor is constructor

    def test_bind_constructor__constructor_required(self):
        binder = Binder()
        self.assertRaisesRegex(InjectorException, "Constructor cannot be None", binder.bind_to_constructor, int, None)



from dataclasses import dataclass

import inject
from test import BaseTestInject
from typing import ClassVar


class TestInjectAttr(BaseTestInject):
    def test_attr_on_dataclass_class_var_raises_error(self):
        with self.assertRaises(inject.InjectorException):
            @dataclass
            class MyClass:
                field: ClassVar[int] = inject.attr(int)


class TestInjectAttrDataclass(BaseTestInject):
    def test_attr_dc(self):
        @dataclass
        class MyClass(object):
            field = inject.attr_dc(int)

        inject.configure(lambda binder: binder.bind(int, 123))
        my = MyClass()
        value0 = my.field
        value1 = my.field

        assert value0 == 123
        assert value1 == 123

    def test_attr_on_dataclass_class_var_works(self):
        @dataclass
        class MyClass:
            field: ClassVar[int] = inject.attr_dc(int)

        inject.configure(lambda binder: binder.bind(int, 123))

        assert MyClass().field == 123
        assert MyClass.field == 123
